// ----------------------------------------------------------------------
//
// inputTime -- TIming tests Hep3Vector input 
//
// Currently tested:
//
// M. Fischler	    11/19/99	First version:
//				On a 533 MHz PIII, under gcc 3.0.1, no-opt,
//				each read and push_back of a Hep3Vector takes
//				47 usec, compared to 37 in the old (inflexible)
//				code.
//				
// 
// ----------------------------------------------------------------------

#include "CLHEP/Vector/AxisAngle.h"
#include "CLHEP/Vector/EulerAngles.h"
#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/config/TemplateFunctions.h"
#include "CLHEP/config/iostream.h"
#include "CLHEP/config/fstream.h"

//-ap #include "ZMtools/ZMtimer.h"
//-ap simplified version, as we don't have ZOOM available here ...
#include <time.h>
class ZMcpuTimer {

public:
  ZMcpuTimer()  {}
  ~ZMcpuTimer() {}

  void start() {  startTime_ = clock(); }
  void stop () {    
    clock_t stopTime = clock();
    elapsed_ += (stopTime - startTime_) / double( CLOCKS_PER_SEC );
    startTime_ = stopTime;
  }

  double cpuTime() { return elapsed_; }

private:
  clock_t             startTime_;
  double elapsed_;
};


#include <vector>

#ifdef HEP_USE_STD
using HepStd::cin;
using HepStd::cout;
using HepStd::cerr;
using HepStd::endl;
//#ifndef _WIN32
//using HepStd::exp;
//#endif
#endif


// **********************
//
// SECTION I - General tools for the various tests
//
// **********************

// **********************
//
// SECTION II - Preparation 
//
// **********************

void prepare (HepStd::vector<Hep3Vector>  & vv) {
  for ( int i = 0; i < 1000; ++i) {
    vv.push_back (Hep3Vector ( .25, -.5, 2.5 ) );
  }
  return;
}

// **********************
//
// SECTION II - Output
//
// **********************


void output  (HepStd::vector<Hep3Vector>  & vv, HepStd::ofstream & f, int n ) {
  unsigned int i;
  for (int k = 0; k < n; ++k) {
  for ( i=0; i< vv.size(); ++i ) {
    f << vv[i] << "\n";	
  }
  }
}

// **********************************
//
// SECTION III - Timing of Input
//
// *********************************

int input   (HepStd::vector<Hep3Vector>  & vv, HepStd::ifstream & f ) {
  unsigned int k;
  Hep3Vector v;
  for ( k=0; f; ++k ) {
    vv.clear();
    for ( int i=0; i< 1000; ++i ) {
      f >> v;
      if (!f) break;
      vv.push_back (Hep3Vector ( .25, -.5, 2.5 ) );
    }
  }
  return k-1;
}


// **********************
//
// SECTION IV - Main
//
// **********************

int main() {

  cout << "Enter 0 for output or 1 for input: ";
  int io;
  cin >> io;
  HepStd::vector<Hep3Vector> vv;
  if (io) {
    HepStd::ifstream f ("inputTime.out");
    if (!f) {
      cout << "Could not open inputTime.out\n";
      return -1;
    }
    ZMcpuTimer timer;
    timer.start();
    int k = input (vv, f);
    timer.stop();
    cout << "Time is " << timer.cpuTime() << "\n";
    cout << "Time per Hep3Vector input is " << 
		1000.0*timer.cpuTime()/k << " usec\n"; 
  } else {
    prepare (vv);
    HepStd::ofstream f ("inputTime.out");
    cout << "Enter number of 1000-Vector input cycles: ";
    int n;
    cin >> n;
    output(vv, f, n);
  }
  return 0;
}

